int verbosity = 2;
#include <cmath>
#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include "rgraph.hpp"
using namespace std;

#include "throwassert.hpp"
#include "Expression.hpp"
#include "MatriceCreuse_tpl.hpp"
#include "QuadratureFormular.hpp"
using namespace Fem2D;

void GradGrad(MatriceElementaireSymetrique < R > &mat, const FElement & K,
	      double *p, int, int, void *)
{				// 
	const Triangle & T = K.T;	// the triangle
	R *aa = mat.a;
	for (int i = 0; i < 3; i++) {
		Vertex & vi = T[i];
		R2 Hi = T.H(i);	// some optimisation 
		for (int j = 0; j <= i; j++) {
			Vertex & vj = T[j];
			R2 Hj = T.H(j);	// some optimisation 
			*aa++ = (Hi, Hj) * T.area;
	}}
}

R g(R2 & P)
{
	return 3 * P.x + 5 * P.y;
}

R f(R2 & P)
{
	return 1;
}

int main(int argc, char **argv)
{
	cout << " begin" << endl;
	initgraphique();
	const char *fn = (argc < 2) ? "c30x30.msh" : argv[1];
	Mesh Th(fn);
	Th.renum();
	Th.Draw();
	FESpace Vh(Th);
	MatriceProfile < R > A(Vh);
	MatriceElementaireSymetrique < R > mate(Vh);
	mate.element = GradGrad;
	for (int i = 0; i < Th.nt; i++)
		A += mate(i, -1, 0, 0);
	const R tgv = 1e30;
	RN U(Th.nv);
	U = 0;

	for (int i = 0; i < Th.nt; i++) {
		// set the  current Intergration formular
		const QuadratureFormular & FI(QuadratureFormular_T_2);
		assert(FI.on == 3);	// verification 
		const Triangle & K(Th[i]);
		const FElement S(Vh[i]);
		R area = K.area;
		int i0 = S[0];	// the 3 vertex number
		int i1 = S[1];
		int i2 = S[2];
		for (int j = 0; j < FI.n; j++)	// loop on integration's points
		{
			QuadraturePoint PI = FI[j];
			// the value of the 3 basics function 
			R w1(PI.x), w2(PI.y), w0(1 - w1 - w2);
			R2 P = K[0] * w0 + K[1] * w1 + K[2] * w2;
			R F = f(P) * PI.a;
			U[i0] += F * w0 * area;
			U[i1] += F * w1 * area;
			U[i2] += F * w2 * area;
		}
	}
	for (int i = 0; i < Th.nv; i++)
		if (!!(Label) Th(i))	// just for boundary  vertex
			U[i] = g(Th(i)) * (A.D[i] = tgv);
	A.cholesky();
	U /= A;

	RN Viso(20);
	Th.InitDraw();
	SetDefaultIsoValue(U, Viso);
	reffecran();
	openPS("laplace");
	PlotValue(Viso, 0, "Isovalue");
	Vh.Draw(U, Viso);
	closePS();
	rattente(1);
	closegraphique();

	return 0;
}
