//Fichier Laplace.cpp modifie par Alain Perronnet
// a partir de l'original ecrit par Frederic HECHT  fem-v5/Laplace0.cpp
//--------------------------------------------------------------------------------

//Resolution du Laplacien sur un carre: Solution exacte U(x,y)=3*x+5*y

//--------------------------------------------------------------------------------
//Les fichiers include

int verbosity = 2;		//Definit le niveau des impressions

#include <cmath>
#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>		//pour lire le fichier de maillage

#include "rgraph.hpp"		//pour les traces
using namespace std;

#include "throwassert.hpp"
#include "Expression.hpp"
#include "MatriceCreuse_tpl.hpp"	//pour la matrice globale EF profil
#include "QuadratureFormular.hpp"	//pour les formules d'integration numerique

using namespace Fem2D;


//---------------------------------------------------------------------------------
//Construction de la matrice elementaire Ae sur l'element K

void GradGrad(MatriceElementaireSymetrique < R > &Ae, const FElement & K, double *p, int, int, void *)	//3 parametres inutilises ici ie,label,stack
	       //(GradGrad est un pointeur de fonction)
{
	const Triangle & T = K.T;	//le triangle a traiter

	R *aa = Ae.a;		//pointeur sur le coefficient courant de la matrice elementaire

	for (int i = 0; i < 3; i++) {
		Vertex & vi = T[i];
		R2 Hi = T.H(i);	//optimisation: [DPe]i = grad lambda^e_i = Hauteur
		for (int j = 0; j <= i; j++) {
			Vertex & vj = T[j];
			R2 Hj = T.H(j);	//optimisation: [DPe]j = grad lambda^e_j
			*aa++ = (Hi, Hj) * T.area;	//operator , du produit scalaire dans R2
		}
	}
}


//---------------------------------------------------------------------------------
//Donnees de U Dirichlet (Solution exacte) et F Omega

R UDirichlet(R2 & P)
{
	return 3 * P.x + 5 * P.y;
}				//Condition de Dirichlet et solution

R fOmega(R2 & P)
{
	return 1;
}				//Source sur Omega


//----------------------------------------------------------------------------------
//Le programme principal

int main(int argc, char **argv)
{
	cout << "Debut de resolution du Laplacien sur le carre unite" <<
	    endl;
	initgraphique();	//grep -in initgraphique *.[hc]pp   pour trouver sa definition
	//=> rgraph.hpp et Xrgraph.cpp 
	//trace avec X11-Windows System du MIT

	const char *fn = (argc < 2) ? "c30x30.msh" : argv[1];	//Le nom du fichier de maillage
	//peut etre le 2-eme mot de la commande Exemple:  Laplace MonMaillage.msh

	Mesh Th(fn);		//grep -in 'class Mesh' *.hpp  => fem.hpp  pour trouver sa definition
	//=> RefCounter.hpp (mutable <=> ne peut jamais etre const)
	//constructeur du maillage par lecture d'un fichier de nom fn
	//construction d'un tableau de nt triangles, nv vertices
	//et neb edges on boundary
	Th.renum();		//renumerotation des noeuds pour reduire le profil de la matrice globale
	Th.Draw();		//Trace du maillage
	rattente(1);		//Pour voir le maillage avec un clic pour continuer

	FESpace Vh(Th);		//constructeur FESpace(const Mesh & TTh) grep => FESpace.hpp et .cpp
	//Definition de l'espace d'interpolation: par defaut triangle P1

	MatriceProfile < R > Ag(Vh);	//Constructeur de la matrice globale
	MatriceElementaireSymetrique < R > Ae(Vh);	//Constructeur de la matrice elementaire

	Ae.element = GradGrad;	//Pointeur de fonction
	for (int i = 0; i < Th.nt; i++)	//Assemblage de Ae du triangle i dans la matrice globale 
		Ag += Ae(i, -1, 0, 0);	//+= operator de MatriceCreuse
	//() operator de MatriceElementaire(k,ie,label,s)
	//<=>call(k,ie,label,s)

	RN Ug(Th.nv);		//Tableau de taille le nombre de vertices
	Ug = 0;			//Mise a zero

	for (int i = 0; i < Th.nt; i++)	//Calcul du second membre global
	{			//boucle sur les triangles
		// set the current Integration formular
		const QuadratureFormular & FI(QuadratureFormular_T_2);	//cf FormuleIntegration.cpp
		assert(FI.on == 3);	// verification 

		const Triangle & K(Th[i]);
		const FElement S(Vh[i]);
		R area = K.area;

		int i0 = S[0];	//the 3 vertex number of the triangle i
		int i1 = S[1];
		int i2 = S[2];

		for (int j = 0; j < FI.n; j++)	//loop on integration's points
		{
			QuadraturePoint PI = FI[j];	//XY of the integration point j 
			R w1(PI.x), w2(PI.y), w0(1 - w1 - w2);	//the value of the 3 basics function

			R2 P = K[0] * w0 + K[1] * w1 + K[2] * w2;

			R F = fOmega(P) * PI.a;

			Ug[i0] += F * w0 * area;	//vecteur global partie interne
			Ug[i1] += F * w1 * area;
			Ug[i2] += F * w2 * area;
		}
	}

	const R tgv = 1e30;
	for (int i = 0; i < Th.nv; i++)	//Traitement des noeuds sur la frontiere Dirichlet
		if (!!(Label) Th(i))	//just for boundary Dirichlet vertex 
			Ug[i] = UDirichlet(Th(i)) * (Ag.D[i] = tgv);

	Ag.cholesky();		// Ag = L tL      Factorisation de la matrice globale apres C.L.
	Ug /= Ag;		// Ug = Ag**-1 Ug Resolution du systeme lineaire

	//Traces graphiques
	RN Viso(20);
	Th.InitDraw();
	SetDefaultIsoValue(Ug, Viso);
	reffecran();
	openPS("laplace");
	PlotValue(Viso, 0, "Isovalue");
	Vh.Draw(Ug, Viso);
	closePS();
	rattente(1);
	closegraphique();

	return 0;
}
