int verbosity = 2;
#include <cmath>
#include <cstdlib>
#include <cstring>
#include <cassert>
#include <iostream>
#include <fstream>
#include "rgraph.hpp"
using namespace std;

#include "throwassert.hpp"
#include "Expression.hpp"
#include "MatriceCreuse_tpl.hpp"
#include "QuadratureFormular.hpp"
using namespace Fem2D;

void GradGrad(MatriceElementaireSymetrique < R > &mat, const FElement & FK,
	      double *p, int, int, void *mydata)
{				// 
	R alpha = 0;
	R beta = 1;
	if (mydata) {
		pair < double, double >*ab =
		    static_cast < pair < double, double >*>(mydata);
		alpha = ab->first;
		beta = ab->second;
	}

	const QuadratureFormular & QF(QuadratureFormular_T_5);
	const Triangle & K = FK.T;	// the triangle
	bool whatd[last_operatortype];
	initwhatd(whatd, 0);
	whatd[1] = whatd[2] = true;
	int nbDoF = FK.NbDoF();
	KNMK_ < R > w(p, nbDoF, FK.N, last_operatortype);
	for (int ipq = 0; ipq < QF.n; ipq++) {
		QuadraturePoint Pq(QF[ipq]);
		R cgg = beta * K.area * Pq.a;
		R cmm = alpha * K.area * Pq.a;
		FK.BF(whatd, Pq, w);
		R *aa = mat.a;
		for (int i = 0; i < nbDoF; i++) {
			for (int j = 0; j <= i; j++) {	//w(i,j,k)=fonction de base i pour inc variationnelle j 
				//et k ordre derivation  0 valeur, 1 DX, 2 DY
				*aa++ +=
				    (w(i, 0, 1) * w(j, 0, 1) +
				     w(i, 0, 2) * w(j, 0, 2)) * cgg + (w(i,
									 0,
									 0)
								       *
								       w(j,
									 0,
									 0))
				    * cmm;
		}}
	}
}

R g(const R2 & P)
{
	return P.x * P.x + P.y * P.y;
}

R f(const R2 & P)
{
	return -4;
}

void Interpole(const FESpace & Vh, KN_ < R > &U, R(*g[])(const R2 & P))
{				// interpole le tableau de pointeur sur de fonction g dans U.
	const Mesh & Th = Vh.Th;
	const int N = Vh.N;
	FElement::aIPJ ipj(Vh[0].Pi_h_ipj());	// index de la matrice d'interpolation  
	FElement::aR2 PtHat(Vh[0].Pi_h_R2());	// les points d'interpolation    
	KN < R > Aipj(ipj.N());	//  pour le coef de la matrice d'interpolation
	KNM < R > Vp(N, PtHat.N());	//  pour stocker les valeur de la fonction au point
	RN ghK(Vh.MaximalNbOfDF() * 3 * Vh.N);	// pour stocker l'interpoler sur l'element

	// pour toutes les elements finis

	for (int k = 0; k < Th.nt; k++) {
		const FElement K(Vh[k]);	// l'elemet fini
		const Triangle & T(K.T);	// le triangle

		K.Pi_h(Aipj);

		for (int p = 0; p < PtHat.N(); p++)	// pour tout les points
			for (int composante = 0; composante < N; composante++)	// pour toutes des composanets      
				Vp(composante, p) =
				    g[composante] (T(PtHat[p]));

		ghK = 0;
		for (int i = 0; i < Aipj.N(); i++) {
			const FElement::IPJ & ipj_i(ipj[i]);
			ghK[ipj_i.i] += Aipj[i] * Vp(ipj_i.j, ipj_i.p);
		}

		for (int dfl = 0; dfl < K.NbDoF(); dfl++) {
			int df = K(dfl);
			U[df] = ghK(dfl);
		}
	}
}


int main(int argc, char **argv)
{
	cout << " begin" << endl;
	initgraphique();
	const char *fn = (argc < 2) ? "c30x30.msh" : argv[1];
	Mesh Th(fn);
	R2 Pmin, Pmax;
	Th.BoundingBox(Pmin, Pmax);
	cout << " Bounding Box " << Pmin << " " << Pmax << endl;
	Th.renum();
	Th.Draw();
	TypeOfFE *TFE = &P1Lagrange;
	const char *nameFE = (argc < 2) ? "P1" : argv[2];	//si pas 2-eme parametre P1 par defaut
	if (strcmp("P1", nameFE) == 0) {
		TFE = &P1Lagrange;
	} else if (strcmp("P2", nameFE) == 0) {
		TFE = &P2Lagrange;
	} else if (strcmp("P1nc", nameFE) == 0) {
		TFE = &P1ncLagrange;
	} else {
		cout << "Error  Element Fini inconnue" << nameFE << endl;
		return (1);
	}

	cout << " L'element Fini : " << nameFE << endl;
	FESpace Vh(Th, *TFE);
	const int NbOfDF = Vh.NbOfDF;	// nb de degre de liberte
	Vh.renum();
	MatriceProfile < R > A(Vh);
	MatriceElementaireSymetrique < R > mate(Vh);

	mate.element = GradGrad;	//assemblage de la matrice globale
	for (int i = 0; i < Th.nt; i++)
		A += mate(i, -1, 0, 0);

	RN U(NbOfDF), Uexact(NbOfDF);	//Vecteur solution calculee et exacte
	U = 0;

	R(**tabg) (const R2 &) = new(R(*[Vh.N]) (const R2 &));
	tabg[0] = g;
	cout << "xx\n";
	cout << " " << tabg[0] (R2(0, 0)) << endl;
	cout << "yy\n";

	//Assemblage du vecteur second membre global
	Interpole(Vh, Uexact, tabg);	//transforme une fonction g(R2) en un tableau[NbOfDF] => "interpolation Lagrange"

	{
		RNMK vxy(Vh.MaximalNbOfDF(), Vh.N, 3);	//stockage des valeurs des fonctions de base et derivees
		RN_ wi(vxy('.', 0, 0));	//selectionne la valeur de la composante 0 des "inconnues variationnelles"

		bool whatd[last_operatortype];
		initwhatd(whatd, 0);	// on ne calcule que les valeurs des fonctions de bases 
		for (int i = 0; i < Th.nt; i++)	//sur les Triangles
		{
			// set the  current Integration formular
			const QuadratureFormular &
			    QF(QuadratureFormular_T_5);
			assert(QF.on == 3);	// verification triangle
			const FElement FK(Vh[i]);	//construction de l'element fini i
			const Triangle & K(FK.T);	//recuperation de la geometrie
			R area = K.area;
			for (int j = 0; j < QF.n; j++)	// loop on integration's points
			{
				QuadraturePoint QP = QF[j];	//R2 + poids dans le triangle de reference
				R2 P(K(QP));	//XY du point d'integration dans le triangle K
				R cF = f(P) * QP.a * area;
				FK.BF(whatd, P, vxy);	//recuperation fonctions de base au pt d'integration dans vxy(nbofd,nbinc,10)
				for (int kk = 0; kk < FK.NbDoF(); kk++)
					U[FK(kk)] += cF * wi[kk];	//valeur des fonctions de base au point d'integration numerique
			}
		}
		//Destruction des variables locales au bloc
	}

	{

		//condition Dirichlet non homogene U=g sur Gamma
		const R tgv = 1e30;
		const int N = Vh.N;
		FElement::aIPJ ipj(Vh[0].Pi_h_ipj());	// index de la matrice d'interpolation  
		FElement::aR2 PtHat(Vh[0].Pi_h_R2());	// les points d'interpolation    
		KN < R > Aipj(ipj.N());	//  pour le coef de la matrice d'interpolation
		KNM < R > Vp(N, PtHat.N());	//  pour stocker les valeur de la fonction au point
		RN ghK(Vh.MaximalNbOfDF() * 3 * Vh.N);	// pour stocker l'interpoler sur l'element


		for (int ib = 0; ib < Th.neb; ib++)	// pour toutes les aretes frontieres 
		{
			int ie, k = Th.BoundaryTriangle(ib, ie);	//ie no de l'arete frontiere ib  du triangle k
			const FElement K(Vh[k]);	// l'elemet fini
			const Triangle & T(K.T);	// le triangle
			int label = Th.bedges[ib].lab;

			K.Pi_h(Aipj);

			for (int p = 0; p < PtHat.N(); p++)	// pour tout les points
				for (int composante = 0; composante < N; composante++)	// pour toutes des composanets      
					Vp(composante, p) = g(T(PtHat[p]));

			ghK = 0;
			for (int i = 0; i < Aipj.N(); i++) {
				const FElement::IPJ & ipj_i(ipj[i]);
				ghK[ipj_i.i] += Aipj[i] * Vp(ipj_i.j, ipj_i.p);	//coefficient de la fonction de base de l'element
			}

			for (int dfl = 0; dfl < K.NbDoF(); dfl++)
				if (onWhatIsEdge[ie][K.DFOnWhat(dfl)]) {	// le DL dfl est sur l'arete frontiere ie du triangle T
					int df = K(dfl);	//no dof local =>global
//              cout << df << " " << dfl << " " << ghK(dfl) << endl;
					U[df] = ghK(dfl) * (A.D[df] = tgv);
				}
		}
	}

	A.cholesky();
	U /= A;
	cout << " U =min " << U.min() << " max =" << U.max() << endl;
	RN Viso(20);
	Th.InitDraw();
	SetDefaultIsoValue(U, Viso);
	reffecran();
	openPS("laplace.eps");
	PlotValue(Viso, 0, "Isovalue");
	Vh.Draw(U, Viso);
	Vh.Draw(Uexact, Viso);
	closePS();
	rattente(1);
	closegraphique();

	return 0;
}
