//Fichier Laplace.cpp modifie par Alain Perronnet
// a partir de l'original ecrit par Frederic HECHT  fem-v5/Laplace1.cpp
//------------------------------------------------------------------------------

//Resolution du Laplacien : Solution exacte U(x,y)=3*x+5*y

//------------------------------------------------------------------------------

int verbosity = 2;		//Definit le niveau des impressions

#include <cmath>
#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>		//pour lire le fichier de maillage

#include "rgraph.hpp"		//pour les traces
using namespace std;

#include "throwassert.hpp"
#include "Expression.hpp"
#include "MatriceCreuse_tpl.hpp"	//pour la matrice globale EF profil
#include "QuadratureFormular.hpp"	//pour les formules d'integration numerique

using namespace Fem2D;

/*
 * Labels:
 */
#if 0
//  testfrontiere.msh : 
//      1   {est le NUMERO ou LABEL dans LIGNE      de } C1                      
//      2   {est le NUMERO ou LABEL dans LIGNE      de } C2                      
// que du robin
#define isNeumann( x ) ((x)==2)
#define isDirichlet( x ) ((x)==1)
#define MESHFILE "testfrontiere.msh"

#else
//  testfrontiere_big.msh : 
//    3   {est le NUMERO ou LABEL dans LIGNE      de } C1B
//    4   {est le NUMERO ou LABEL dans LIGNE      de } C2B
#define isDirichlet( x ) ((x)==3)
#define isNeumann(  x ) ((x)==4)
#define MESHFILE "testfrontiere_big.msh"
#endif

/*
 * Donnees du tenseur de conductivite
 * tenseurConductivite[i] est la diagonale du tenseur pour
 * le label "i"
 */
#define MaxLabel 4
R tenseurConductivite[][2] = {
	{1., 1.},		// idendite, le plus classique
	{1., 1.},		// idendite, le plus classique
	{1., 1.},		// idendite, le plus classique
	{.5, 2.},		// surtout sur l'axe des y
	{2., .5},		// surtout sur l'axe des x
	{0., 1.},		// rien selon l'ace des x
};

//------------------------------------------------------------------------------
//Construction de la matrice elementaire Ae sur l'element K

void GradGrad(MatriceElementaireSymetrique < R > &Ae, const FElement & K, double *p, int, int, void *)	//3 parametres inutilises ici ie,label,stack
	       //(GradGrad est un pointeur de fonction)
{
	const Triangle & T = K.T;	//le triangle a traiter

	R *aa = Ae.a;		//pointeur sur le coefficient courant de la matrice elementaire

	for (int i = 0; i < 3; i++) {
		R2 vi = (R2) T[i];
		R2 Hj;		// optimisation: on declare ici avant la boucle
		assert(((Label) T) < MaxLabel);	// check for tenseurConductivite[][]
		R2 gmi(T.H(i).x * tenseurConductivite[((Label) T).lab][0],	// gradient modifie = A(x)*grad(wi)
		       T.H(i).y * tenseurConductivite[((Label) T).lab][1]);	// ou A(x) est diagonale
		for (int j = 0; j <= i; j++) {
			Hj = T.H(j);	//optimisation: [DPe]j = grad lambda^e_j
			*aa++ = (gmi, Hj) * T.area;	//operator , du produit scalaire dans R2
		}
	}
}

//------------------------------------------------------------------------------
//
// on prend l'exemple de teta(x,y) = x*x + y*y
#define rayon1 ( 4. )
#define rayon2 ( 10. )
R TetaExact(R2 & P)
{
	return P.x * P.x + P.y * P.y;
}

R UDirichlet(R2 & P)
{
	return rayon1 * rayon1;
}

R UNeumann(R2 & P)
{
	return rayon2 * 2;
}

R fOmega(R2 & P)
{
	return -4.;
}				// laplacien constant = -4

//------------------------------------------------------------------------------
//Le programme principal

int main(int argc, char **argv)
{
	cout << "Debut de resolution du Laplacien sur le carre unite" <<
	    endl;
	initgraphique();	//grep -in initgraphique *.[hc]pp   pour trouver sa definition
	//=> rgraph.hpp et Xrgraph.cpp 
	//trace avec X11-Windows System du MIT

	//Le nom du fichier de maillage
	const char *fn = (argc < 2) ? MESHFILE : argv[1];
	//peut etre le 2-eme mot de la commande Exemple:  Laplace MonMaillage.msh

	Mesh Th(fn);		//grep -in 'class Mesh' *.hpp => fem.hpp pour trouver definition
	//=> RefCounter.hpp (mutable <=> ne peut jamais etre const)
	//constructeur du maillage par lecture d'un fichier de nom fn
	//construction d'un tableau de nt triangles, nv vertices
	//et neb edges on boundary
	Th.renum();		//renumerotation des noeuds pour reduire le profil de la matrice
	Th.Draw();		//Trace du maillage
	rattente(1);		//Pour voir le maillage avec un clic pour continuer

	FESpace Vh(Th);		//constructeur FESpace(const Mesh & TTh) grep => FESpace.*pp
	//Definition de l'espace d'interpolation: par defaut triangleP1

	MatriceProfile < R > Ag(Vh);	//Constructeur de la matrice globale
	MatriceElementaireSymetrique < R > Ae(Vh);	//Constructeur matrice elementaire

	Ae.element = GradGrad;	//Pointeur de fonction
	for (int i = 0; i < Th.nt; i++)	//Assemblage de Ae du triangle i dans lamatrice Ag
		Ag += Ae(i, -1, 0, 0);	//+= operator de MatriceCreuse
	//() operator de MatriceElementaire(k,ie,label,s)
	//<=>call(k,ie,label,s)

	RN Ug(Th.nv);		//Tableau de taille le nombre de vertices
	Ug = 0;			//Mise a zero

	for (int i = 0; i < Th.nt; i++)	//Calcul du second membre global
	{			//boucle sur les triangles
		// set the current Integration formular cf FormuleIntegration.cpp
		const QuadratureFormular & FI(QuadratureFormular_T_2);
		assert(FI.on == 3);	// verification 3 points d'integration

		const Triangle & K(Th[i]);	//Le triangle i
		const FElement S(Vh[i]);	//Les 3 sommets du triangle i
		R area = K.area;

		int i0 = S[0];	//the 3 vertex number of the triangle i
		int i1 = S[1];
		int i2 = S[2];

		for (int j = 0; j < FI.n; j++)	//loop on integration's points
		{
			QuadraturePoint PI = FI[j];	//XY of the integration point j 
			R w1(PI.x), w2(PI.y), w0(1 - w1 - w2);	//the value of the 3 basics function

			R2 P = K[0] * w0 + K[1] * w1 + K[2] * w2;

			R F = fOmega(P) * PI.a;

			Ug[i0] += F * w0 * area;	//vecteur global partie interne
			Ug[i1] += F * w1 * area;	//assemble a partir du vecteur elementaire
			Ug[i2] += F * w2 * area;
		}
	}

	const R tgv = 1e30;
	for (int i = 0; i < Th.nv; i++)	//Traitement des noeuds sur la frontiere Dirichlet
		if (isDirichlet(((Label) Th(i)).lab))
			Ug[i] = UDirichlet(Th(i)) * (Ag.D[i] = tgv);

	// participation dus aux conditions de type Neumann
	for (int i = 0; i < Th.neb; i++)	//boucle sur les cotes (boundary edges)
	{
		const BoundaryEdge & edge(Th.bedges[i]);
		R length = edge.length();

		if (!isNeumann(((Label) edge).lab))
			continue;	// not neumann -> continue the loop

		// set the current Integration formular cf FormuleIntegration.cpp
		const QuadratureFormular1d & FI(QF_GaussLegendre3);
		assert(FI.n == 3);	// verification 3 points d'integration

		// find out vertex indices.. didn't find a better way with present fem API
		int i0 = &(edge[0]) - Th.vertices;
		assert(i0 >= 0);
		assert(i0 < Th.nv);
		int i1 = &(edge[1]) - Th.vertices;
		assert(i1 >= 0);
		assert(i1 < Th.nv);


		for (int j = 0; j < FI.n; j++)	//loop on integration's points
		{
			QuadratureFormular1d::Point PI = FI[j];	// position/poids de du point j

			R w0(PI.x), w1(1 - PI.x);	// valeurs des fonctions chapeaux
			R2 P = ((R2) edge[0]) * w0 + ((R2) edge[1]) * w1;	// point courant
			R N = UNeumann(P) * PI.p;	// contribution de ce point

			Ug[i0] += N * w0 * length;	//vecteur global partie interne
			Ug[i1] += N * w1 * length;	//assemble a partir du vecteur elementaire
		}
	}

	Ag.cholesky();		// Ag = L tL   Factorisation de la matrice globale apres C.L.
	Ug /= Ag;		// Ug = Ag**-1 Ug Resolution du systeme lineaire

	//Trace graphique de la temperature sous forme d'isothermes
	RN Viso(20);
	Th.InitDraw();
	SetDefaultIsoValue(Ug, Viso);
	reffecran();
	openPS("laplace");
	PlotValue(Viso, 0, "Isovalue");
	Vh.Draw(Ug, Viso);
	closePS();
	rattente(1);
	closegraphique();

	// test de validite
	R max = 0;
	assert(Ug.N() == Th.nv);
	for (int i = 0; i < Th.nv; i++)	// on parcours l'ensemble des points
	{
		R diff;
		R2 P;

		P = (R2) (Th.vertices[i]);
		diff = Ug[i] - TetaExact(P);
		diff *= diff;	// on prend le carre
		if (diff > max)
			max = diff;
	}
	printf("\nMax des carres obtenu : %f\n\n", max);

	return 0;
}
