int verbosity = 2;		//Definit le niveau des impressions

#include <cmath>
#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>		//pour lire le fichier de maillage

#include "rgraph.hpp"		//pour les traces
using namespace std;

#include "throwassert.hpp"
#include "Expression.hpp"
#include "MatriceCreuse_tpl.hpp"	//pour la matrice globale EF profil
#include "QuadratureFormular.hpp"	//pour les formules d'integration numerique

using namespace Fem2D;

/*
 * Labels:
 */

// choose one :

#define USEROBIN
//#undef USEROBIN

// from piece.msh / piecerobin.msh
//      21   {est le NUMERO ou LABEL dans LIGNE      de } DIRICHLET               
//      23   {est le NUMERO ou LABEL dans LIGNE      de } NEUMANNINT              
//      22   {est le NUMERO ou LABEL dans LIGNE      de } NEUMANNEXT              
//      33   {est le NUMERO ou LABEL dans LIGNE      de } RAD                     
//
#ifdef USEROBIN
// avec condition de robin
#define isDirichlet( x ) ((x)==21 || (x)==23)
#define isRobin( x ) ( (x)==33)
#define isNeumann( x ) ( (x)==22 )
#define MESHFILE "piecerobin.msh"
#else
// sans condition de robin
#define isDirichlet( x ) ((x)==21 || (x)==23)
#define isRobin( x ) (false)
#define isNeumann( x ) ( (x)==22 )
//#define isNeumann( x ) ( (x)==23 || (x)==22 )
#define MESHFILE "piece.msh"
#endif

/*
 * Donnees du tenseur de conductivite
 * tenseurConductivite[i] est la diagonale du tenseur pour
 * le label "i"
 */
// from piece.msh
//       2   {est le NUMERO dans SURFACE    de }MUR1                    
//       3   {est le NUMERO dans SURFACE    de }MUR2                    
//       6   {est le NUMERO dans SURFACE    de }MUR3                    
//       5   {est le NUMERO dans SURFACE    de }SURFINT                 
//       7   {est le NUMERO dans SURFACE    de }SURFINTRAD              

#define MaxLabel 7
#define ConductiviteMur 0.05
#define ConductiviteAir 2.
R tenseurConductivite[][2] = {
	{ConductiviteAir, ConductiviteAir},	// ne devrait pas etre utilise // 0
	{ConductiviteAir, ConductiviteAir},	// ne devrait pas etre utilise // 1
	{ConductiviteMur, ConductiviteMur},	// mur 1; 
	{ConductiviteMur, ConductiviteMur},	// mur 2; 
	{ConductiviteAir, ConductiviteAir},	// ne devrait pas etre utilise // 4
	{ConductiviteAir, ConductiviteAir},	// Surfint
	{ConductiviteMur, ConductiviteMur},	// mur 3; 
	{ConductiviteAir, ConductiviteAir},	// Surfintrad (avec radiateur) // 7
};

//------------------------------------------------------------------------------
//Construction de la matrice elementaire Ae sur l'element K

void GradGrad(MatriceElementaireSymetrique < R > &Ae, const FElement & K, double *p, int, int, void *)	//3 parametres inutilises ici ie,label,stack
	       //(GradGrad est un pointeur de fonction)
{
	const Triangle & T = K.T;	//le triangle a traiter

	R *aa = Ae.a;		//pointeur sur le coefficient courant de la matrice elementaire

	for (int i = 0; i < 3; i++) {
		R2 vi = (R2) T[i];
		R2 Hj;		// optimisation: on declare ici avant la boucle
		assert(((Label) T).lab <= MaxLabel);	// check for tenseurConductivite[][]
		R2 gmi(T.H(i).x * tenseurConductivite[((Label) T).lab][0],	// gradient modifie = A(x)*grad(wi)
		       T.H(i).y * tenseurConductivite[((Label) T).lab][1]);	// ou A(x) est diagonale
		for (int j = 0; j <= i; j++) {
			Hj = T.H(j);	//optimisation: [DPe]j = grad lambda^e_j
			*aa++ = (gmi, Hj) * T.area;	//operator , du produit scalaire dans R2
		}
	}
}

//------------------------------------------------------------------------------

R UDirichlet(R2 & P, Label label)
{
	switch (label.lab) {
	default:
		assert(false);
	case 23:
		return 20;	// murs interieurs
	case 21:
		return 0;	// murs exterieurs
	}
}
R UNeumann(R2 & P, Label label)
{
	switch (label.lab) {
	default:
		assert(false);
	case 23:
		return 0;	// NEUMANNINT : pas de chaleur traversant les murs interieurs
	case 22:
		return .5;	// NEUMANNEXT le soleil chauffe la piece
	}
}
R fOmega(R2 & P)
{
	return 0;
}				// Pas de creation de chaleur a l'interieur

#ifdef USEROBIN
R GRobin(R2 & P)
{
	return 0.5;
}				// coefficient d'echange pour la frontiere de Robin

R TetaExtRobin(R2 & P)
{
	return 60;
}				// temperature interieur du radiateur
#endif

//------------------------------------------------------------------------------
//Le programme principal

int main(int argc, char **argv)
{
	cout << "Debut de resolution du Laplacien sur le carre unite" <<
	    endl;
	initgraphique();	//grep -in initgraphique *.[hc]pp   pour trouver sa definition
	//=> rgraph.hpp et Xrgraph.cpp 
	//trace avec X11-Windows System du MIT

	//Le nom du fichier de maillage
	const char *fn = (argc < 2) ? MESHFILE : argv[1];
	//peut etre le 2-eme mot de la commande Exemple:  Laplace MonMaillage.msh

	Mesh Th(fn);		//grep -in 'class Mesh' *.hpp => fem.hpp pour trouver definition
	//=> RefCounter.hpp (mutable <=> ne peut jamais etre const)
	//constructeur du maillage par lecture d'un fichier de nom fn
	//construction d'un tableau de nt triangles, nv vertices
	//et neb edges on boundary
	Th.renum();		//renumerotation des noeuds pour reduire le profil de la matrice
	Th.Draw();		//Trace du maillage
	rattente(1);		//Pour voir le maillage avec un clic pour continuer

	FESpace Vh(Th);		//constructeur FESpace(const Mesh & TTh) grep => FESpace.*pp
	//Definition de l'espace d'interpolation: par defaut triangleP1

	MatriceProfile < R > Ag(Vh);	//Constructeur de la matrice globale
	MatriceElementaireSymetrique < R > Ae(Vh);	//Constructeur matrice elementaire

	Ae.element = GradGrad;	//Pointeur de fonction
	for (int i = 0; i < Th.nt; i++)	//Assemblage de Ae du triangle i dans lamatrice Ag
		Ag += Ae(i, -1, 0, 0);	//+= operator de MatriceCreuse
	//() operator de MatriceElementaire(k,ie,label,s)
	//<=>call(k,ie,label,s)

	RN Ug(Th.nv);		//Tableau de taille le nombre de vertices
	Ug = 0;			//Mise a zero

	for (int i = 0; i < Th.nt; i++)	//Calcul du second membre global
	{			//boucle sur les triangles
		// set the current Integration formular cf FormuleIntegration.cpp
		const QuadratureFormular & FI(QuadratureFormular_T_2);
		assert(FI.on == 3);	// verification 3 points d'integration

		const Triangle & K(Th[i]);	//Le triangle i
		const FElement S(Vh[i]);	//Les 3 sommets du triangle i
		R area = K.area;

		int i0 = S[0];	//the 3 vertex number of the triangle i
		int i1 = S[1];
		int i2 = S[2];

		for (int j = 0; j < FI.n; j++)	//loop on integration's points
		{
			QuadraturePoint PI = FI[j];	//XY of the integration point j 
			R w1(PI.x), w2(PI.y), w0(1 - w1 - w2);	//the value of the 3 basics function

			R2 P = K[0] * w0 + K[1] * w1 + K[2] * w2;

			R F = fOmega(P) * PI.a * area;

			Ug[i0] += F * w0;	//vecteur global partie interne
			Ug[i1] += F * w1;	//assemble a partir du vecteur elementaire
			Ug[i2] += F * w2;
		}
	}

	// set the current Integration formular cf FormuleIntegration.cpp
	const QuadratureFormular1d & FI(QF_GaussLegendre3);
	assert(FI.n == 3);	// verification 3 points d'integration

	/* 
	 * Frontiere de Neumann
	 */
	for (int i = 0; i < Th.neb; i++)	//boucle sur les cotes (boundary edges)
	{
		const BoundaryEdge & edge(Th.bedges[i]);
		R length = edge.length();
		Label label = ((Label) edge).lab;

		if (!isNeumann(label))
			continue;

		// find out vertex indices.. didn't find a better way with present fem API
		int i0 = &(edge[0]) - Th.vertices;
		assert(i0 >= 0);
		assert(i0 < Th.nv);
		int i1 = &(edge[1]) - Th.vertices;
		assert(i1 >= 0);
		assert(i1 < Th.nv);


		for (int j = 0; j < FI.n; j++)	//loop on integration's points
		{
			QuadratureFormular1d::Point PI = FI[j];	// position/poids de du point j

			R w0(PI.x), w1(1 - PI.x);	// valeurs des fonctions chapeaux
			R2 P = ((R2) edge[0]) * w0 + ((R2) edge[1]) * w1;	// point courant

			R N = length * UNeumann(P, label) * PI.p;	// contribution de ce point

			Ug[i0] += N * w0;	//vecteur global partie interne
			Ug[i1] += N * w1;	//assemble a partir du vecteur elementaire
		}
	}

#ifdef USEROBIN
	/* 
	 * Frontiere de Robin
	 */
	for (int i = 0; i < Th.neb; i++)	//boucle sur les cotes (boundary edges)
	{
		const BoundaryEdge & edge(Th.bedges[i]);
		R length = edge.length();
		Label label = ((Label) edge).lab;
		R2 P;

		if (!isRobin(label))
			continue;

		// find out vertex indices.. didn't find a better way with present fem API
		int i0 = &(edge[0]) - Th.vertices;
		assert(i0 >= 0);
		assert(i0 < Th.nv);
		int i1 = &(edge[1]) - Th.vertices;
		assert(i1 >= 0);
		assert(i1 < Th.nv);


		// pas besoin de formule d'integration pour la matrice
		P = (R2) (edge[1] + edge[0]) / 2;
		R asdf = GRobin(P) * length / 6;
		Ag.add2x2Matrice(i0, i1, asdf * 2., asdf, asdf * .2);

		for (int j = 0; j < FI.n; j++)	//loop on integration's points
		{
			R N;
			QuadratureFormular1d::Point PI = FI[j];	// position/poids de du point j

			R w0(PI.x), w1(1 - PI.x);	// valeurs des fonctions chapeaux
			P = ((R2) edge[1]) * w0 + ((R2) edge[0]) * w1;	// point courant

			// modification du vecteur Ug
			N = TetaExtRobin(P) * GRobin(P) * length * PI.p;	// contribution de ce point
			Ug[i0] += N * w0;
			Ug[i1] += N * w1;
		}
	}
#endif

	/* 
	 * Frontiere de Dirichlet
	 */
	const R tgv = 1e30;
	for (int i = 0; i < Th.nv; i++)	//Traitement des noeuds sur la frontiere Dirichlet
		if (isDirichlet(((Label) Th(i)).lab))
			Ug[i] =
			    UDirichlet(Th(i), (Label) Th(i)) * (Ag.D[i] =
								tgv);



	Ag.cholesky();		// Ag = L tL   Factorisation de la matrice globale apres C.L.
	Ug /= Ag;		// Ug = Ag**-1 Ug Resolution du systeme lineaire

	//Trace graphique de la temperature sous forme d'isothermes
	RN Viso(20);
	Th.InitDraw();
	SetDefaultIsoValue(Ug, Viso);
	reffecran();
	openPS("laplace");
	PlotValue(Viso, 0, "Isovalue");
	Vh.Draw(Ug, Viso);
	closePS();
	rattente(1);
	closegraphique();


	return 0;
}
